import logging
import os
from pathlib import Path

logger = logging.getLogger(__name__)

from moat.util import yload


def load_cfg(cfg):  # pylint: disable=redefined-outer-name
    cfg = Path(cfg).absolute()
    if cfg.exists():
        pass
    elif (ct := cfg.parent/"tests"/cfg.name).exists():  # pragma: no cover
        cfg = ct
    elif (ct := cfg.parent.parent/cfg.name).exists():  # pragma: no cover
        cfg = ct
    else:  # pragma: no cover
        raise RuntimeError(f"Config file {cfg!r} not found")

    with cfg.open("r", encoding="utf-8") as f:
        cfg = yload(f)

    from logging.config import dictConfig

    cfg["disable_existing_loggers"] = False
    try:
        dictConfig(cfg)
    except ValueError:
        pass
    logging.captureWarnings(True)
    logger.debug("Test %s", "starting up")
    return cfg


cfg = load_cfg(os.environ.get("LOG_CFG", "logging.cfg"))


import trio._core._run as tcr

if "PYTHONHASHSEED" in os.environ:
    tcr._ALLOW_DETERMINISTIC_SCHEDULING = True
    tcr._r.seed(os.environ["PYTHONHASHSEED"])



"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.
"""

# compatibility

from moat.util import *  # noqa: F403

__all__ = ["packer", "unpacker", "stream_unpacker"]  # noqa: F405



import logging

import pytest
import trio
from asyncscope import scope
from moat.util import P, PathLongener
from moat.src.test import raises

from moat.kv.auth import loader
from moat.kv.client import ServerError
from moat.kv.mock.mqtt import stdtest

logger = logging.getLogger(__name__)


async def collect(i, path=()):
    res = []
    pl = PathLongener(path)
    async for r in i:
        r.pop("tock", 0)
        r.pop("seq", 0)
        pl(r)
        res.append(r)
    return res


@pytest.mark.trio
async def test_71_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=100) as st:
        assert st is not None
        async with st.client() as c:
            await c._request(
                "set_internal",
                path=P("codec.int"),
                value={
                    "in": [("1", 1), ("2", 2), ("3", 3)],
                    "out": [(1, "1"), (2, "2"), (-3, "-3")],
                    "encode": "return str(value)",
                    "decode": "assert isinstance(value,str); return int(value)",
                },
            )
            await c._request("set_internal", path=P("conv.foo.inty.#"), value={"codec": "int"})
            um = loader("_test", "user", make=True, server=False)
            u = um.build({"name": "std"})
            await u.send(c)
            u = um.build({"name": "con"})
            await c._request(
                "set_internal",
                path=P("auth._test.user.con.conv"),
                value=dict(key="foo"),
                iter=False,
            )
            await u.send(c)
            await c._request("set_auth_typ", typ="_test")

        recv = []
        um = loader("_test", "user", make=False, server=False)

        async def mon(evt):
            async with st.client(auth=um.build({"name": "std"})) as c:
                async with c._stream("watch", path=P("inty")) as q:
                    evt.set()
                    pl = PathLongener(P("inty"))
                    async for m in q:
                        pl(m)
                        del m["tock"]
                        del m["seq"]
                        recv.append(m)

        evt = trio.Event()
        await scope.spawn(mon, evt)
        await evt.wait()
        async with st.client(auth=um.build({"name": "con"})) as c:
            await c.set(P("inty.ten"), value="10")
            with raises(ServerError):
                await c.set(P("inty.nope"), value=11)
            await c.set(P("inty.yep.yepyepyep"), value="13")
            with raises(ServerError):
                await c.set(P("inty.nope.nopenope"), value=12)
            await c.set(P("inty.yep.yepyepyep.yep"), value="99")
            await c.set(P("inty"), value="hello")

            r = await c.get(P("inty"))
            assert r.value == "hello"
            r = await c.get(P("inty.ten"))
            assert r.value == "10"
            r = await c.get(P("inty.yep.yepyepyep"))
            assert r.value == "13"

            # run_c = partial(run, "-D", "kv", "-h", s.ports[0][0], "-p", s.ports[0][1])
            # await run_c("-a","_test name=std", "get", "-rd_", do_stdout=False)

        assert recv == [
            {"path": P("inty.ten"), "value": 10},
            {"path": P("inty.yep.yepyepyep"), "value": 13},
            {"path": P("inty.yep.yepyepyep.yep"), "value": 99},
            {"path": P("inty"), "value": "hello"},
        ]



import io
import logging
from functools import partial

import pytest
from moat.util import P, PathLongener
from moat.src.test import raises

from moat.kv.client import ServerError
from moat.kv.mock import run
from moat.kv.mock.mqtt import stdtest

logger = logging.getLogger(__name__)


async def collect(i, path=()):
    res = []
    pl = PathLongener(path)
    async for r in i:
        r.pop("tock", 0)
        r.pop("seq", 0)
        pl(r)
        res.append(r)
    return res


@pytest.mark.trio
async def test_71_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=80) as st:
        assert st is not None
        async with st.client() as c:
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int"),
                    value={
                        "bad": ["foo", None],
                        "good": [0, 1, 2],
                        "code": "if not isinstance(value,int): rise ValueError('not an int')",
                        # yes this checks for the typo (SyntaxError=
                    },
                )
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int"),
                    value={
                        "bad": ["foo", None],
                        "good": [0, 1, "dud"],
                        "code": "if not isinstance(value,int): raise ValueError('not an int')",
                    },
                )
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int"),
                    value={
                        "bad": ["foo", 4],
                        "good": [0, 1, 2],
                        "code": "if not isinstance(value,int): raise ValueError('not an int')",
                    },
                )
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int"),
                    value={
                        "bad": [],
                        "good": [0, 1, 2],
                        "code": "if not isinstance(value,int): raise ValueError('not an int')",
                    },
                )
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int"),
                    value={
                        "bad": ["foo", None],
                        "good": [1],
                        "code": "if not isinstance(value,int): raise ValueError('not an int')",
                    },
                )
            await c._request(
                "set_internal",
                path=P("type.int"),
                value={
                    "bad": ["foo", None],
                    "good": [0, 1, 2],
                    "code": "if not isinstance(value,int): raise ValueError('not an int')",
                },
            )
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int.percent"),
                    value={
                        "bad": ["fuf", 101],
                        "good": [0, 55, 100],
                        "code": "if not 0<=value<=100: raise ValueError('not a percentage')",
                    },
                )
            with raises(ServerError):
                await c._request(
                    "set_internal",
                    path=P("type.int.percent"),
                    value={
                        "bad": ["fuf", 101],
                        "good": [0, 5.5, 100],
                        "code": "if not 0<=value<=100: raise ValueError('not a percentage')",
                    },
                )
            await c._request(
                "set_internal",
                path=P("type.int.percent"),
                value={
                    "bad": [-1, 101],
                    "good": [0, 55, 100],
                    "code": "if not 0<=value<=100: raise ValueError('not a percentage')",
                },
            )
            with raises(ServerError):
                await c._request(
                    "set_internal", path=P("match.one.+.two"), value={"tope": P("int.percent")}
                )
            await c._request(
                "set_internal", path=P("match.one.+.two"), value={"type": P("int.percent")}
            )

            await c.set(P("one.x.two"), value=99)
            with raises(ServerError):
                await c.set(P("one.y.two"), value=9.9)
            with raises(ServerError):
                await c.set(P("one.y.two"), value="zoz")
            await c.set(P("one.y"), value="zoz")

            pass  # client end
        pass  # server end


@pytest.mark.trio
async def test_72_cmd(autojump_clock, tmpdir):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=80) as st:
        assert st is not None
        (s,) = st.s
        async with st.client() as c:
            h = p = None  # pylint
            for h, p, *_ in s.ports:
                if h[0] != ":":
                    break
            rr = partial(run, "kv", "-h", h, "-p", p, do_stdout=False)
            path = tmpdir.join("foo")
            with io.open(path, "w", encoding="utf-8") as f:
                f.write(
                    """\
good:
- 0
- 2
bad:
- none
- "Foo"
code: "if not isinstance(value,int): raise ValueError('not an int')"
"""
                )
            await rr("type", "set", "-d", str(path), "int")

            with io.open(path, "w", encoding="utf-8") as f:
                f.write("if not 0<=value<=100: raise ValueError('not a percentage')\n")

            with raises(ServerError):
                await rr(
                    "type",
                    "set",
                    "-s",
                    str(path),
                    "-g",
                    "0",
                    "-g",
                    "100",
                    "-g",
                    "50",
                    "-b",
                    "-1",
                    "-b",
                    "5.5",
                    "int.percent",
                )
            await rr(
                "type",
                "set",
                "-s",
                str(path),
                "-g",
                "0",
                "-g",
                "100",
                "-g",
                "50",
                "-b",
                "-1",
                "-b",
                "555",
                "int.percent",
            )

            await rr("type", "match", "-t", "int.percent", "foo.+.bar")

            with raises(ServerError):
                await rr("data", "foo.dud.bar", "set", "-v", ":", "123")
            with raises(ServerError):
                await rr("data", "foo.dud.bar", "set", "-e", ":", "123")
            with raises(ServerError):
                await rr("data", "foo.dud.bar", "set", "-e", ":", "5.5")
            await rr("data", "foo.dud.bar", "set", "-e", ":", "55")

            assert (await c.get(P("foo.dud.bar"))).value == 55

            pass  # client end
        pass  # server end
